@@ -1,5 +1,5 @@
 use super::{BracketPair, Class, Level, OpeningBracket};
-use crate::{util, Error};
+use crate::Error;
 use std::{
   cmp::{max, min},
   ops::{Index, IndexMut, Range},
@@ -52,13 +52,13 @@ impl IsolatingRunSequence {
           },
           Class::AL => processing_classes[i] = Class::R,
           _ => {},
-        };
+        }
 
         match w2_processing_class {
           Class::L | Class::R => last_strong_is_al = false,
           Class::AL => last_strong_is_al = true,
           _ => {},
-        };
+        }
 
         let class_before_w456 = processing_classes[i];
 
@@ -122,7 +122,7 @@ impl IsolatingRunSequence {
             },
           },
           _ => {},
-        };
+        }
 
         bn_run_indices.clear();
         prev_class_before_w5 = processing_classes[i];
@@ -149,7 +149,7 @@ impl IsolatingRunSequence {
         Class::L => last_strong_is_l = true,
         Class::R | Class::AL => last_strong_is_l = false,
         _ => {},
-      };
+      }
     }
   }
 
@@ -157,12 +157,12 @@ impl IsolatingRunSequence {
     &self,
     text: &str,
     original_classes: &[Class],
-  ) -> Vec<BracketPair> {
-    let mut ret = Vec::new();
+    bracket_pairs: &mut Vec<BracketPair>,
+  ) {
     let mut stack = Vec::new();
 
     for (run_index, level_run) in self.runs.iter().enumerate() {
-      for (i, ch) in util::sliced(text, level_run.clone()).char_indices() {
+      for (i, ch) in text[level_run.clone()].char_indices() {
         let actual_index = level_run.start + i;
 
         if original_classes[actual_index] != Class::ON {
@@ -182,7 +182,7 @@ impl IsolatingRunSequence {
             .rev()
             .find(|(_, element)| element.0 == matched.opening)
           {
-            ret.push(BracketPair {
+            bracket_pairs.push(BracketPair {
               start: element.1,
               end: actual_index,
               start_run: element.2,
@@ -195,8 +195,7 @@ impl IsolatingRunSequence {
       }
     }
 
-    ret.sort_by_key(|r| r.start);
-    ret
+    bracket_pairs.sort_by_key(|r| r.start);
   }
 
   pub(crate) fn resolve_implicit_neutral(
@@ -208,17 +207,16 @@ impl IsolatingRunSequence {
     let e = levels[self.runs[0].start].class();
 
     let not_e = if e == Class::L { Class::R } else { Class::L };
+    let mut bracket_pairs = Vec::new();
 
-    for pair in self.identify_bracket_pairs(text, processing_classes) {
+    self.identify_bracket_pairs(text, processing_classes, &mut bracket_pairs);
+
+    for pair in bracket_pairs {
       let mut found_e = false;
       let mut found_not_e = false;
       let mut class_to_set = None;
 
-      let start_char_len = util::sliced(text, pair.start..)
-        .chars()
-        .next()
-        .unwrap()
-        .len_utf8();
+      let start_char_len = text[pair.start..].chars().next().unwrap().len_utf8();
 
       for enclosed_i in self.iter_forwards_from(pair.start + start_char_len, pair.start_run) {
         if enclosed_i >= pair.end {
@@ -259,11 +257,7 @@ impl IsolatingRunSequence {
       }
 
       if let Some(class_to_set) = class_to_set {
-        let end_char_len = util::sliced(text, pair.end..)
-          .chars()
-          .next()
-          .unwrap()
-          .len_utf8();
+        let end_char_len = text[pair.end..].chars().next().unwrap().len_utf8();
 
         for class in
           (pair.start..pair.start + start_char_len).chain(pair.end..pair.end + end_char_len)
@@ -326,7 +320,7 @@ impl IsolatingRunSequence {
               next_class = self.end_class;
               break;
             },
-          };
+          }
         }
 
         let new_class = match (prev_class, next_class) {
@@ -372,6 +366,7 @@ pub(crate) struct Paragraph {
   pub(crate) range: Range<usize>,
   pub(crate) level: Level,
   pub(crate) pure_ltr: bool,
+  pub(crate) has_isolate_controls: bool,
 }
 
 impl Paragraph {
@@ -380,7 +375,7 @@ impl Paragraph {
     &'a self,
     slicable: &'a T,
   ) -> &'a <T as Index<Range<usize>>>::Output {
-    util::sliced(slicable, self.range.clone())
+    &slicable[self.range.clone()]
   }
 
   #[inline(always)]
@@ -388,7 +383,7 @@ impl Paragraph {
     &'a self,
     slicable: &'a mut T,
   ) -> &'a mut <T as Index<Range<usize>>>::Output {
-    util::sliced_mut(slicable, self.range.clone())
+    &mut slicable[self.range.clone()]
   }
 
   pub(crate) fn visual_runs(
@@ -509,7 +504,8 @@ impl Paragraph {
     original_classes: &[Class],
     processing_classes: &mut [Class],
     levels: &mut [Level],
-  ) {
+    runs: &mut Vec<Range<usize>>,
+  ) -> Result<(), Error> {
     let mut stack = vec![Status {
       level: self.level,
       status: OverrideStatus::Neutral,
@@ -519,6 +515,9 @@ impl Paragraph {
     let mut overflow_embedding_count = 0;
     let mut valid_isolate_count = 0;
 
+    let mut current_run_level = Level::ltr();
+    let mut current_run_start = 0;
+
     for (idx, character) in input.char_indices() {
       let current_class = original_classes[idx];
 
@@ -530,7 +529,10 @@ impl Paragraph {
         | Class::RLI
         | Class::LRI
         | Class::FSI => {
-          let last = stack.last().unwrap();
+          let Some(last) = stack.last() else {
+            return Err(Error::MalformedOverrideStatusStack);
+          };
+
           levels[idx] = last.level;
 
           let is_isolate = current_class.is_isolate();
@@ -595,7 +597,10 @@ impl Paragraph {
             valid_isolate_count -= 1;
           }
 
-          let last = stack.last().unwrap();
+          let Some(last) = stack.last() else {
+            return Err(Error::MalformedOverrideStatusStack);
+          };
+
           levels[idx] = last.level;
 
           match last.status {
@@ -609,19 +614,32 @@ impl Paragraph {
           if overflow_isolate_count <= 0 {
             if overflow_embedding_count > 0 {
               overflow_embedding_count -= 1;
-            } else if stack.len() >= 2 && stack.last().unwrap().status != OverrideStatus::Isolate {
-              stack.pop();
+            } else {
+              let Some(last) = stack.last() else {
+                return Err(Error::MalformedOverrideStatusStack);
+              };
+
+              if stack.len() >= 2 && last.status != OverrideStatus::Isolate {
+                stack.pop();
+              }
             }
           }
 
-          levels[idx] = stack.last().unwrap().level;
+          let Some(last) = stack.last() else {
+            return Err(Error::MalformedOverrideStatusStack);
+          };
+
+          levels[idx] = last.level;
           processing_classes[idx] = Class::BN;
         },
 
         Class::B => {},
 
         _ => {
-          let last = stack.last().unwrap();
+          let Some(last) = stack.last() else {
+            return Err(Error::MalformedOverrideStatusStack);
+          };
+
           levels[idx] = last.level;
 
           if current_class != Class::BN {
@@ -638,112 +656,176 @@ impl Paragraph {
         levels[idx + j] = levels[idx];
         processing_classes[idx + j] = processing_classes[idx];
       }
-    }
-  }
 
-  pub(crate) fn isolating_run_sequences<'a>(
-    &'a self,
-    levels: &'a [Level],
-    original_classes: &'a [Class],
-  ) -> impl Iterator<Item = IsolatingRunSequence> + 'a {
-    let mut runs = Vec::new();
-
-    if !levels.is_empty() {
-      let mut current_run_level = levels[0];
-      let mut current_run_start = 0;
-
-      for i in 1..levels.len() {
-        if !original_classes[i].removed_by_x9() && levels[i] != current_run_level {
-          runs.push(current_run_start..i);
-
-          current_run_level = levels[i];
-          current_run_start = i;
-        }
+      if idx == 0 {
+        current_run_level = levels[idx];
+      } else if original_classes[idx].removed_by_x9() && levels[idx] != current_run_level {
+        runs.push(current_run_start..idx);
+        current_run_level = levels[idx];
+        current_run_start = idx;
       }
+    }
 
+    if levels.len() > current_run_start {
       runs.push(current_run_start..levels.len());
     }
 
-    let mut sequences = Vec::with_capacity(runs.len());
-    let mut stack = vec![vec![]];
+    Ok(())
+  }
 
-    for run in runs {
-      let start_class = original_classes[run.start];
-      let end_class = original_classes[run.clone()]
-        .iter()
-        .copied()
-        .rev()
-        .find(|x| !x.removed_by_x9())
-        .unwrap_or(start_class);
+  pub(crate) fn isolating_run_sequences(
+    &self,
+    levels: &[Level],
+    level_runs: &[Range<usize>],
+    original_classes: &[Class],
+    irs: &mut Vec<IsolatingRunSequence>,
+  ) -> Result<(), Error> {
+    if self.has_isolate_controls {
+      let mut runs = Vec::new();
 
-      let mut sequence = if start_class == Class::PDI && stack.len() > 1 {
-        stack.pop().unwrap()
-      } else {
-        Vec::with_capacity(1)
-      };
+      if let Some(&(mut current_run_level)) = levels.first() {
+        let mut current_run_start = 0;
 
-      sequence.push(run);
+        for i in 1..levels.len() {
+          if !original_classes[i].removed_by_x9() && levels[i] != current_run_level {
+            runs.push(current_run_start..i);
 
-      if end_class.is_isolate() {
-        stack.push(sequence);
-      } else {
-        sequences.push(sequence);
+            current_run_level = levels[i];
+            current_run_start = i;
+          }
+        }
+
+        runs.push(current_run_start..levels.len());
+      }
+
+      let mut sequences = Vec::with_capacity(runs.len());
+      let mut stack = vec![vec![]];
+
+      for run in runs {
+        let start_class = original_classes[run.start];
+        let end_class = original_classes[run.clone()]
+          .iter()
+          .copied()
+          .rev()
+          .find(|x| !x.removed_by_x9())
+          .unwrap_or(start_class);
+
+        let mut sequence = if start_class == Class::PDI && stack.len() > 1 {
+          stack.pop().unwrap()
+        } else {
+          Vec::with_capacity(1)
+        };
+
+        sequence.push(run);
+
+        if end_class.is_isolate() {
+          stack.push(sequence);
+        } else {
+          sequences.push(sequence);
+        }
+      }
+
+      sequences.extend(stack.into_iter().rev().filter(|seq| !seq.is_empty()));
+
+      irs.reserve_exact(sequences.len());
+
+      for sequence in sequences {
+        if sequence.is_empty() {
+          return Err(Error::MalformedIsolatingRunSequence);
+        }
+
+        let sequence_start = sequence[0].start;
+        let runs_len = sequence.len();
+        let sequence_end = sequence[runs_len - 1].end;
+
+        let mut result = IsolatingRunSequence {
+          runs: sequence,
+          start_class: Class::L,
+          end_class: Class::L,
+        };
+
+        let sequence_level = levels[result
+          .iter_forwards_from(sequence_start, 0)
+          .find(|&i| !original_classes[i].removed_by_x9())
+          .unwrap_or(sequence_start)];
+
+        let end_level = levels[result
+          .iter_backwards_from(sequence_end, runs_len - 1)
+          .find(|&i| !original_classes[i].removed_by_x9())
+          .unwrap_or(sequence_end - 1)];
+
+        let preceeding_level = match original_classes[..sequence_start]
+          .iter()
+          .rposition(|x| !x.removed_by_x9())
+        {
+          Some(idx) => levels[idx],
+          None => self.level,
+        };
+
+        let last_non_removed = original_classes[..sequence_end]
+          .iter()
+          .copied()
+          .rev()
+          .find(|x| !x.removed_by_x9())
+          .unwrap_or(Class::BN);
+
+        let succeeding_level = if last_non_removed.is_isolate() {
+          self.level
+        } else {
+          match original_classes[sequence_end..]
+            .iter()
+            .position(|x| !x.removed_by_x9())
+          {
+            Some(idx) => levels[sequence_end + idx],
+            None => self.level,
+          }
+        };
+
+        result.start_class = max(sequence_level, preceeding_level).class();
+        result.end_class = max(end_level, succeeding_level).class();
+
+        irs.push(result);
+      }
+    } else {
+      irs.reserve_exact(level_runs.len());
+
+      for run in level_runs {
+        let run_levels = &levels[run.clone()];
+        let run_classes = &original_classes[run.clone()];
+        let seq_level = run_levels[run_classes
+          .iter()
+          .position(|c| !c.removed_by_x9())
+          .unwrap_or(0)];
+
+        let end_level = run_levels[run_classes
+          .iter()
+          .rposition(|c| !c.removed_by_x9())
+          .unwrap_or(run.end - run.start - 1)];
+
+        let pred_level = match original_classes[..run.start]
+          .iter()
+          .rposition(|c| !c.removed_by_x9())
+        {
+          Some(idx) => levels[idx],
+          None => self.level,
+        };
+
+        let succ_level = match original_classes[run.end..]
+          .iter()
+          .position(|c| !c.removed_by_x9())
+        {
+          Some(idx) => levels[run.end + idx],
+          None => self.level,
+        };
+
+        irs.push(IsolatingRunSequence {
+          runs: vec![run.clone()],
+          start_class: max(seq_level, pred_level).class(),
+          end_class: max(end_level, succ_level).class(),
+        });
       }
     }
 
-    sequences.extend(stack.into_iter().rev().filter(|seq| !seq.is_empty()));
-
-    sequences.into_iter().map(move |sequence| {
-      let mut result = IsolatingRunSequence {
-        runs: sequence,
-        start_class: Class::L,
-        end_class: Class::L,
-      };
-
-      let sequence_start = result.runs[0].start;
-      let runs_len = result.runs.len();
-      let sequence_end = result.runs[runs_len - 1].end;
-
-      let sequence_level = levels[result
-        .iter_forwards_from(sequence_start, 0)
-        .find(|i| !original_classes[*i].removed_by_x9())
-        .unwrap_or(sequence_start)];
-
-      let end_level = levels[result
-        .iter_backwards_from(sequence_end, runs_len - 1)
-        .find(|i| !original_classes[*i].removed_by_x9())
-        .unwrap_or(sequence_end - 1)];
-
-      let preceeding_level = match original_classes[..sequence_start]
-        .iter()
-        .rposition(|x| !x.removed_by_x9())
-      {
-        Some(idx) => levels[idx],
-        None => self.level,
-      };
-
-      let last_non_removed = original_classes[..sequence_end]
-        .iter()
-        .copied()
-        .rev()
-        .find(|x| !x.removed_by_x9())
-        .unwrap_or(Class::BN);
-
-      let succeeding_level = if last_non_removed.is_isolate() {
-        self.level
-      } else {
-        match original_classes[sequence_end..]
-          .iter()
-          .position(|x| !x.removed_by_x9())
-        {
-          Some(idx) => levels[sequence_end + idx],
-          None => self.level,
-        }
-      };
-
-      result.start_class = max(sequence_level, preceeding_level).class();
-      result.end_class = max(end_level, succeeding_level).class();
-      result
-    })
+    Ok(())
   }
 }
