@@ -3,6 +3,8 @@
 mod filter;
 mod filter_map;
 mod head;
+mod skip;
 mod sort;
 mod sort_by;
 mod sort_by_key;
+mod tail;
