@@ -2,17 +2,18 @@
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-use std::mem::{replace, swap, MaybeUninit};
+use std::mem::{replace, swap};
 use std::ops::{Range, RangeBounds};
 use std::ptr::null;
 use std::sync::atomic::{AtomicPtr, Ordering};
-use std::{mem, ptr};
+
+use archery::{SharedPointer, SharedPointerKind};
 
 use crate::nodes::chunk::Chunk;
 use crate::sync::Lock;
-use crate::util::{to_range, PoolRef, Ref};
+use crate::util::to_range;
 use crate::vector::{
-    Iter, IterMut, RRBPool, Vector,
+    GenericVector, Iter, IterMut,
     VectorInner::{Full, Inline, Single},
     RRB,
 };
@@ -96,27 +97,27 @@ fn check_indices<const N: usize>(len: usize, indices: &[usize; N]) -> Option<()>
 /// [Iter]: struct.Iter.html
 /// [narrow]: #method.narrow
 /// [split_at]: #method.split_at
-pub enum Focus<'a, A> {
+pub enum Focus<'a, A, P: SharedPointerKind> {
     #[doc(hidden)]
     /// The Single variant is a focus of a simple Vector that can be represented as a single slice.
     Single(&'a [A]),
     #[doc(hidden)]
     /// The Full variant is a focus of a more complex Vector that cannot be represented as a single slice.
-    Full(TreeFocus<A>),
+    Full(TreeFocus<A, P>),
 }
 
-impl<'a, A> Focus<'a, A>
+impl<'a, A, P: SharedPointerKind> Focus<'a, A, P>
 where
     A: 'a,
 {
     /// Construct a `Focus` for a [`Vector`][Vector].
     ///
     /// [Vector]: enum.Vector.html
-    pub fn new(vector: &'a Vector<A>) -> Self {
+    pub fn new(vector: &'a GenericVector<A, P>) -> Self {
         match &vector.vector {
-            Inline(_, chunk) => Focus::Single(chunk),
-            Single(_, chunk) => Focus::Single(chunk),
-            Full(_, tree) => Focus::Full(TreeFocus::new(tree)),
+            Inline(chunk) => Focus::Single(chunk),
+            Single(chunk) => Focus::Single(chunk),
+            Full(tree) => Focus::Full(TreeFocus::new(tree)),
         }
     }
 
@@ -180,9 +181,9 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let vec = Vector::from_iter(0..1000);
+    /// let vec: Vector<i64> = Vector::from_iter(0..1000);
     /// let narrowed = vec.focus().narrow(100..200);
-    /// let narrowed_vec = narrowed.into_iter().cloned().collect();
+    /// let narrowed_vec: Vector<i64> = narrowed.into_iter().cloned().collect();
     /// assert_eq!(Vector::from_iter(100..200), narrowed_vec);
     /// ```
     ///
@@ -221,10 +222,10 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let vec = Vector::from_iter(0..1000);
+    /// let vec: Vector<i64> = Vector::from_iter(0..1000);
     /// let (left, right) = vec.focus().split_at(500);
-    /// let left_vec = left.into_iter().cloned().collect();
-    /// let right_vec = right.into_iter().cloned().collect();
+    /// let left_vec: Vector<i64> = left.into_iter().cloned().collect();
+    /// let right_vec: Vector<i64> = right.into_iter().cloned().collect();
     /// assert_eq!(Vector::from_iter(0..500), left_vec);
     /// assert_eq!(Vector::from_iter(500..1000), right_vec);
     /// ```
@@ -248,19 +249,19 @@ where
     }
 }
 
-impl<'a, A> IntoIterator for Focus<'a, A>
+impl<'a, A, P: SharedPointerKind + 'a> IntoIterator for Focus<'a, A, P>
 where
     A: Clone + 'a,
 {
     type Item = &'a A;
-    type IntoIter = Iter<'a, A>;
+    type IntoIter = Iter<'a, A, P>;
 
     fn into_iter(self) -> Self::IntoIter {
         Iter::from_focus(self)
     }
 }
 
-impl<'a, A> Clone for Focus<'a, A>
+impl<'a, A, P: SharedPointerKind> Clone for Focus<'a, A, P>
 where
     A: Clone + 'a,
 {
@@ -272,10 +273,10 @@ where
     }
 }
 
-pub struct TreeFocus<A> {
+pub struct TreeFocus<A, P: SharedPointerKind> {
     /// A clone of the Vector's internal tree that this focus points to. A clone ensures that we don't require a
     /// reference to the original tree.
-    tree: RRB<A>,
+    tree: RRB<A, P>,
     /// The view represents the range of the tree that this TreeFocus can see. The view can be narrowed by calling
     /// either the narrow or split_at methods.
     view: Range<usize>,
@@ -289,7 +290,7 @@ pub struct TreeFocus<A> {
     target_ptr: *const Chunk<A>,
 }
 
-impl<A> Clone for TreeFocus<A> {
+impl<A, P: SharedPointerKind> Clone for TreeFocus<A, P> {
     fn clone(&self) -> Self {
         let tree = self.tree.clone();
         TreeFocus {
@@ -302,17 +303,17 @@ impl<A> Clone for TreeFocus<A> {
     }
 }
 
-unsafe impl<A: Send> Send for TreeFocus<A> {}
-unsafe impl<A: Sync> Sync for TreeFocus<A> {}
+unsafe impl<A: Send, P: SharedPointerKind + Send> Send for TreeFocus<A, P> {}
+unsafe impl<A: Sync, P: SharedPointerKind + Sync> Sync for TreeFocus<A, P> {}
 
 #[inline]
 fn contains<A: Ord>(range: &Range<A>, index: &A) -> bool {
     *index >= range.start && *index < range.end
 }
 
-impl<A> TreeFocus<A> {
+impl<A, P: SharedPointerKind> TreeFocus<A, P> {
     /// Creates a new TreeFocus for a Vector's RRB tree.
-    fn new(tree: &RRB<A>) -> Self {
+    fn new(tree: &RRB<A, P>) -> Self {
         let middle_start = tree.outer_f.len() + tree.inner_f.len();
         let middle_end = middle_start + tree.middle.len();
         TreeFocus {
@@ -445,7 +446,7 @@ impl<A> TreeFocus<A> {
 /// # #[macro_use] extern crate imbl;
 /// # use imbl::vector::Vector;
 /// # use std::iter::FromIterator;
-/// let mut vec = Vector::from_iter(0..1000);
+/// let mut vec: Vector<i64> = Vector::from_iter(0..1000);
 /// let focus1 = vec.focus_mut();
 /// // Fails here in 2015 edition because you're creating
 /// // two mutable references to the same thing.
@@ -462,7 +463,7 @@ impl<A> TreeFocus<A> {
 /// # #[macro_use] extern crate imbl;
 /// # use imbl::vector::Vector;
 /// # use std::iter::FromIterator;
-/// let mut vec = Vector::from_iter(0..1000);
+/// let mut vec: Vector<i64> = Vector::from_iter(0..1000);
 /// let focus = vec.focus_mut();
 /// let (mut left, mut right) = focus.split_at(500);
 /// assert_eq!(Some(&0), left.get(0));
@@ -476,7 +477,7 @@ impl<A> TreeFocus<A> {
 /// # #[macro_use] extern crate imbl;
 /// # use imbl::vector::Vector;
 /// # use std::iter::FromIterator;
-/// let mut vec = Vector::from_iter(0..1000);
+/// let mut vec: Vector<i64> = Vector::from_iter(0..1000);
 /// let (left, right) = {
 ///     let focus = vec.focus_mut();
 ///     focus.split_at(500)
@@ -488,24 +489,24 @@ impl<A> TreeFocus<A> {
 /// ```
 ///
 /// [Focus]: enum.Focus.html
-pub enum FocusMut<'a, A> {
+pub enum FocusMut<'a, A, P: SharedPointerKind> {
     #[doc(hidden)]
     /// The Single variant is a focusmut of a simple Vector that can be represented as a single slice.
-    Single(RRBPool<A>, &'a mut [A]),
+    Single(&'a mut [A]),
     #[doc(hidden)]
     /// The Full variant is a focus of a more complex Vector that cannot be represented as a single slice.
-    Full(RRBPool<A>, TreeFocusMut<'a, A>),
+    Full(TreeFocusMut<'a, A, P>),
 }
 
-impl<'a, A> FocusMut<'a, A>
+impl<'a, A, P: SharedPointerKind> FocusMut<'a, A, P>
 where
     A: 'a,
 {
     /// Get the length of the focused `Vector`.
     pub fn len(&self) -> usize {
         match self {
-            FocusMut::Single(_, chunk) => chunk.len(),
-            FocusMut::Full(_, tree) => tree.len(),
+            FocusMut::Single(chunk) => chunk.len(),
+            FocusMut::Full(tree) => tree.len(),
         }
     }
 
@@ -527,9 +528,9 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let mut vec = Vector::from_iter(0..1000);
+    /// let mut vec: Vector<i64> = Vector::from_iter(0..1000);
     /// let narrowed = vec.focus_mut().narrow(100..200);
-    /// let narrowed_vec = narrowed.unmut().into_iter().cloned().collect();
+    /// let narrowed_vec: Vector<i64> = narrowed.unmut().into_iter().cloned().collect();
     /// assert_eq!(Vector::from_iter(100..200), narrowed_vec);
     /// ```
     ///
@@ -545,8 +546,8 @@ where
             panic!("vector::FocusMut::narrow: range out of bounds");
         }
         match self {
-            FocusMut::Single(pool, chunk) => FocusMut::Single(pool, &mut chunk[r]),
-            FocusMut::Full(pool, tree) => FocusMut::Full(pool, tree.narrow(r)),
+            FocusMut::Single(chunk) => FocusMut::Single(&mut chunk[r]),
+            FocusMut::Full(tree) => FocusMut::Full(tree.narrow(r)),
         }
     }
 
@@ -568,7 +569,7 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let mut vec = Vector::from_iter(0..1000);
+    /// let mut vec: Vector<i64> = Vector::from_iter(0..1000);
     /// {
     ///     let (left, right) = vec.focus_mut().split_at(500);
     ///     for ptr in left {
@@ -591,28 +592,22 @@ where
             panic!("vector::FocusMut::split_at: index out of bounds");
         }
         match self {
-            FocusMut::Single(pool, chunk) => {
+            FocusMut::Single(chunk) => {
                 let (left, right) = chunk.split_at_mut(index);
-                (
-                    FocusMut::Single(pool.clone(), left),
-                    FocusMut::Single(pool, right),
-                )
+                (FocusMut::Single(left), FocusMut::Single(right))
             }
-            FocusMut::Full(pool, tree) => {
+            FocusMut::Full(tree) => {
                 let (left, right) = tree.split_at(index);
-                (
-                    FocusMut::Full(pool.clone(), left),
-                    FocusMut::Full(pool, right),
-                )
+                (FocusMut::Full(left), FocusMut::Full(right))
             }
         }
     }
 
     /// Convert a `FocusMut` into a `Focus`.
-    pub fn unmut(self) -> Focus<'a, A> {
+    pub fn unmut(self) -> Focus<'a, A, P> {
         match self {
-            FocusMut::Single(_, chunk) => Focus::Single(chunk),
-            FocusMut::Full(_, mut tree) => Focus::Full(TreeFocus {
+            FocusMut::Single(chunk) => Focus::Single(chunk),
+            FocusMut::Full(mut tree) => Focus::Full(TreeFocus {
                 tree: {
                     let t = tree.tree.lock().unwrap();
                     (*t).clone()
@@ -626,19 +621,16 @@ where
     }
 }
 
-impl<'a, A> FocusMut<'a, A>
+impl<'a, A, P: SharedPointerKind> FocusMut<'a, A, P>
 where
     A: Clone + 'a,
 {
     /// Construct a `FocusMut` for a `Vector`.
-    pub fn new(vector: &'a mut Vector<A>) -> Self {
+    pub fn new(vector: &'a mut GenericVector<A, P>) -> Self {
         match &mut vector.vector {
-            Inline(pool, chunk) => FocusMut::Single(pool.clone(), chunk),
-            Single(pool, chunk) => FocusMut::Single(
-                pool.clone(),
-                PoolRef::make_mut(&pool.value_pool, chunk).as_mut_slice(),
-            ),
-            Full(pool, tree) => FocusMut::Full(pool.clone(), TreeFocusMut::new(tree)),
+            Inline(chunk) => FocusMut::Single(chunk),
+            Single(chunk) => FocusMut::Single(SharedPointer::make_mut(chunk).as_mut_slice()),
+            Full(tree) => FocusMut::Full(TreeFocusMut::new(tree)),
         }
     }
 
@@ -650,24 +642,25 @@ where
     /// Get a mutable reference to the value at a given index.
     pub fn get_mut(&mut self, index: usize) -> Option<&mut A> {
         match self {
-            FocusMut::Single(_, chunk) => chunk.get_mut(index),
-            FocusMut::Full(pool, tree) => tree.get(pool, index),
+            FocusMut::Single(chunk) => chunk.get_mut(index),
+            FocusMut::Full(tree) => tree.get(index),
         }
     }
 
     fn get_many_mut<const N: usize>(&mut self, indices: [usize; N]) -> Option<[&mut A; N]> {
         check_indices(self.len(), &indices)?;
         match self {
-            FocusMut::Single(_, chunk) => {
+            FocusMut::Single(chunk) => {
                 // FIXME: Stable polyfill for std `get_many_mut`
-                let chunk: *mut A = ptr::from_mut(*chunk).cast();
-                let mut arr = [const { MaybeUninit::uninit() }; N];
-                for idx in 0..N {
-                    arr[idx].write(unsafe { &mut *chunk.add(indices[idx]) });
-                }
-                unsafe { mem::transmute_copy(&arr) }
+                let chunk: *mut A = (*chunk).as_mut_ptr();
+                Some(indices.map(|index| {
+                    // Safety:
+                    // - `check_indices` ensures each index is `< self.len()`, which for `FocusMut::Single` is `chunk.len()`
+                    // - `check_indices` ensures the indexes do not overlap
+                    unsafe { &mut *chunk.add(index) }
+                }))
             }
-            FocusMut::Full(pool, tree) => tree.get_many(pool, indices),
+            FocusMut::Full(tree) => tree.get_many(indices),
         }
     }
 
@@ -727,7 +720,7 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let mut vec = vector![1, 2, 3, 4, 5];
+    /// let mut vec: Vector<i64> = vector![1, 2, 3, 4, 5];
     /// vec.focus_mut().pair(1, 3, |a, b| *a += *b);
     /// assert_eq!(vector![1, 6, 3, 4, 5], vec);
     /// ```
@@ -755,7 +748,7 @@ where
     /// # #[macro_use] extern crate imbl;
     /// # use imbl::vector::Vector;
     /// # use std::iter::FromIterator;
-    /// let mut vec = vector![1, 2, 3, 4, 5];
+    /// let mut vec: Vector<i64> = vector![1, 2, 3, 4, 5];
     /// vec.focus_mut().triplet(0, 2, 4, |a, b, c| *a += *b + *c);
     /// assert_eq!(vector![9, 2, 3, 4, 5], vec);
     /// ```
@@ -780,43 +773,43 @@ where
             panic!("vector::FocusMut::chunk_at: index out of bounds");
         }
         match self {
-            FocusMut::Single(_, chunk) => (0..len, chunk),
-            FocusMut::Full(pool, tree) => {
-                let (range, chunk) = tree.get_chunk(pool, index);
+            FocusMut::Single(chunk) => (0..len, chunk),
+            FocusMut::Full(tree) => {
+                let (range, chunk) = tree.get_chunk(index);
                 (range, chunk)
             }
         }
     }
 }
 
-impl<'a, A> IntoIterator for FocusMut<'a, A>
+impl<'a, A, P: SharedPointerKind> IntoIterator for FocusMut<'a, A, P>
 where
     A: Clone + 'a,
 {
     type Item = &'a mut A;
-    type IntoIter = IterMut<'a, A>;
+    type IntoIter = IterMut<'a, A, P>;
 
     fn into_iter(self) -> Self::IntoIter {
         IterMut::from_focus(self)
     }
 }
 
-impl<'a, A> From<FocusMut<'a, A>> for Focus<'a, A>
+impl<'a, A, P: SharedPointerKind> From<FocusMut<'a, A, P>> for Focus<'a, A, P>
 where
     A: Clone + 'a,
 {
-    fn from(f: FocusMut<'a, A>) -> Focus<'a, A> {
+    fn from(f: FocusMut<'a, A, P>) -> Focus<'a, A, P> {
         f.unmut()
     }
 }
 
 // NOTE: The documentation the mutable version is similar to the non-mutable version. I will comment for the places
 // where there are differences, otherwise the documentation is copied directly.
-pub struct TreeFocusMut<'a, A> {
+pub struct TreeFocusMut<'a, A, P: SharedPointerKind> {
     /// The tree that this TreeFocusMut refers to. Unlike the non-mutable version, TreeFocusMut needs to store a
     /// mutable reference. Additionally, there may be multiple TreeFocusMuts that refer to the same tree so we need a
     /// Lock to synchronise the changes.
-    tree: Lock<&'a mut RRB<A>>,
+    tree: Lock<&'a mut RRB<A, P>>,
     /// The view represents the range of the tree that this TreeFocusMut can see. The view can be narrowed by calling
     /// either the narrow or split_at methods.
     view: Range<usize>,
@@ -832,12 +825,12 @@ pub struct TreeFocusMut<'a, A> {
     target_ptr: AtomicPtr<Chunk<A>>,
 }
 
-impl<'a, A> TreeFocusMut<'a, A>
+impl<'a, A, P: SharedPointerKind> TreeFocusMut<'a, A, P>
 where
     A: 'a,
 {
     /// Creates a new TreeFocusMut for a Vector's RRB tree.
-    fn new(tree: &'a mut RRB<A>) -> Self {
+    fn new(tree: &'a mut RRB<A, P>) -> Self {
         let middle_start = tree.outer_f.len() + tree.inner_f.len();
         let middle_end = middle_start + tree.middle.len();
         TreeFocusMut {
@@ -910,12 +903,12 @@ where
     }
 }
 
-impl<'a, A> TreeFocusMut<'a, A>
+impl<'a, A, P: SharedPointerKind> TreeFocusMut<'a, A, P>
 where
     A: Clone + 'a,
 {
     /// Sets the internal chunk to the one that contains the given absolute index.
-    fn set_focus(&mut self, pool: &RRBPool<A>, index: usize) {
+    fn set_focus(&mut self, index: usize) {
         let mut tree = self
             .tree
             .lock()
@@ -925,13 +918,13 @@ where
             if index < outer_len {
                 self.target_range = 0..outer_len;
                 self.target_ptr.store(
-                    PoolRef::make_mut(&pool.value_pool, &mut tree.outer_f),
+                    SharedPointer::make_mut(&mut tree.outer_f),
                     Ordering::Relaxed,
                 );
             } else {
                 self.target_range = outer_len..self.middle_range.start;
                 self.target_ptr.store(
-                    PoolRef::make_mut(&pool.value_pool, &mut tree.inner_f),
+                    SharedPointer::make_mut(&mut tree.inner_f),
                     Ordering::Relaxed,
                 );
             }
@@ -940,21 +933,21 @@ where
             if index < outer_start {
                 self.target_range = self.middle_range.end..outer_start;
                 self.target_ptr.store(
-                    PoolRef::make_mut(&pool.value_pool, &mut tree.inner_b),
+                    SharedPointer::make_mut(&mut tree.inner_b),
                     Ordering::Relaxed,
                 );
             } else {
                 self.target_range = outer_start..tree.length;
                 self.target_ptr.store(
-                    PoolRef::make_mut(&pool.value_pool, &mut tree.outer_b),
+                    SharedPointer::make_mut(&mut tree.outer_b),
                     Ordering::Relaxed,
                 );
             }
         } else {
             let tree_index = index - self.middle_range.start;
             let level = tree.middle_level;
-            let middle = Ref::make_mut(&mut tree.middle);
-            let (range, ptr) = middle.lookup_chunk_mut(pool, level, 0, tree_index);
+            let middle = SharedPointer::make_mut(&mut tree.middle);
+            let (range, ptr) = middle.lookup_chunk_mut(level, 0, tree_index);
             self.target_range =
                 (range.start + self.middle_range.start)..(range.end + self.middle_range.start);
             self.target_ptr.store(ptr, Ordering::Relaxed);
@@ -962,43 +955,40 @@ where
     }
 
     /// Gets the value at the given index relative to the TreeFocusMut.
-    pub fn get(&mut self, pool: &RRBPool<A>, index: usize) -> Option<&mut A> {
+    pub fn get(&mut self, index: usize) -> Option<&mut A> {
         if index >= self.len() {
             return None;
         }
         let phys_index = self.physical_index(index);
         if !contains(&self.target_range, &phys_index) {
-            self.set_focus(pool, phys_index);
+            self.set_focus(phys_index);
         }
         let target_phys_index = phys_index - self.target_range.start;
         Some(&mut self.get_focus()[target_phys_index])
     }
 
-    fn get_many<const N: usize>(
-        &mut self,
-        pool: &RRBPool<A>,
-        indices: [usize; N],
-    ) -> Option<[&mut A; N]> {
+    fn get_many<const N: usize>(&mut self, indices: [usize; N]) -> Option<[&mut A; N]> {
         check_indices(self.len(), &indices)?;
-        let mut arr = [const { MaybeUninit::uninit() }; N];
-        for idx in 0..N {
-            let phys_idx = self.physical_index(indices[idx]);
+        Some(indices.map(|phys_idx| {
             if !contains(&self.target_range, &phys_idx) {
-                self.set_focus(pool, phys_idx);
+                self.set_focus(phys_idx);
             }
             let target_idx = phys_idx - self.target_range.start;
-            let chunk = self.get_focus_ptr();
-            let ptr = unsafe { Chunk::as_mut_slice_ptr(chunk) };
-            arr[idx].write(unsafe { &mut *ptr.cast::<A>().add(target_idx) });
-        }
-        unsafe { mem::transmute_copy(&arr) }
+            // Safety: we have called `set_focus` to get a valid chunk pointer
+            // and `target_idx` lies within it
+            unsafe {
+                let chunk = self.get_focus_ptr();
+                let ptr: *mut [A] = Chunk::as_mut_slice_ptr(chunk);
+                &mut *ptr.cast::<A>().add(target_idx)
+            }
+        }))
     }
 
     /// Gets the chunk for an index as a slice and its corresponding range within the TreeFocusMut.
-    pub fn get_chunk(&mut self, pool: &RRBPool<A>, index: usize) -> (Range<usize>, &mut [A]) {
+    pub fn get_chunk(&mut self, index: usize) -> (Range<usize>, &mut [A]) {
         let phys_index = self.physical_index(index);
         if !contains(&self.target_range, &phys_index) {
-            self.set_focus(pool, phys_index);
+            self.set_focus(phys_index);
         }
         let mut left = 0;
         let mut right = 0;
